/**
A C program to measure the entropy base on
histrograms

Arguments: file_name length,
where:
	file_name is the name of the file with the total input sequence
	length is the size of the words to calculate their histograms

Fraga 2016
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

int bit[16] = {
0x1, 0x2, 0x4, 0x8,
0x10, 0x20, 0x40, 0x80,
0x100, 0x200, 0x400, 0x800,
0x1000, 0x2000, 0x4000, 0x8000 };

int cadenaAnumero( char *cadena, int m )
{
	int i;
	int n=0;
	char *aux;

	aux = cadena + m - 1;

	for( i=0; i<m; i++ ) {
		if ( *aux == '1' )
			n |= bit[ i ];
		aux--;
	}

	return n;
}


void entropia( char *nombre, int i )
{
	FILE *arch;
	char *buf, *paux, *paux2;
	int *p;
	int tam, j, sec, n, a, m, resto;
	double val, tempo;

	tam = 1<<i;

	p = (int *)malloc( tam*sizeof(int) );
	if ( p == NULL ) {
		fprintf( stderr, "No hay memoria 1\n" );
		exit(3);
	}
	for( j=0; j<tam; j++ )
		p[j] = 0;


	arch = fopen( nombre, "r" );
	if ( arch == NULL ) {
		fprintf( stderr, "No puedo abrir el archivo %s\n", nombre );
		free(p);
		exit(4);
	}

	// Caso 1 
	if ( (80%i)==0 ) {
		buf = (char *)malloc( 82 );
		if ( buf == NULL ){
			fprintf( stderr, "No hay memoria\n" );
			fclose( arch );
			free( p );
			exit(5);
		}	
		while( 1 ) {
			if ( fgets( buf, 82, arch ) == NULL ) 
				break;
			
			n = strnlen(buf,82)/i;
			paux = buf + n;
    		while( *paux == 0 || *paux == '\n' )
				*paux-- = 0;

			n = strnlen(buf,80)/i;
			paux = buf;
			for( j=0; j<n; j++ ) {
				sec = cadenaAnumero( paux, i );
				// printf("%d,", sec );
				p[ sec ]++;
				paux += i;
			}
			// printf("\n");
		}
	}
	else {
		// m = i*(80/i + 1 );
		m = i + 80;
		buf = (char *)malloc( m+2 );
		if ( buf == NULL ){
			fprintf( stderr, "No hay memoria\n" );
			fclose( arch );
			free( p );
			exit(5);
		}	
		paux = buf;
		resto = 0;
		while( 1 ) {
			if ( fgets( paux, 82, arch ) == NULL ) 
				break;
			a = strnlen(buf, m);
			paux = buf + a;
    		while( *paux == 0 || *paux == '\n' )
				*paux-- = 0;

			a = strnlen(buf, m);
			n = a/i;
			// printf("%d %d %s\n", a, n, buf );
			paux2 = buf;
			for( j=0; j<n; j++ ) {
				sec = cadenaAnumero( paux2, i );
				// printf("%d,", sec );
				p[ sec ]++;
				paux2 += i;
			}
			// printf("\n");
			resto = a - n*i;
			paux2 = buf + n*i;
			strncpy( buf, paux2, resto );
			// printf("resto = %d, %s\n", resto, buf);
			paux = buf + resto;
		}
	}
	fclose( arch );

	n = 0;
	for( j=0; j<tam; j++ )
		n += p[ j ];

	val = 0;
	for( j=0; j<tam; j++ ) {
		if ( p[j]>0 ) {
			tempo = (double)p[ j ]/n;
			val -= tempo * log2( tempo );
		}
	}
	val /= i;

	printf( "Size=%d n=%d i=%d H=%lf\n", tam, n, i, val );  

	free( p );
	free( buf );
}

int main( int argc, char *argv[] )
{
	char nombre[32];
	int i;

	if ( argc != 3 ) {
		fprintf( stderr, "Args: file_name lenght\n" );
		return 1;
	}

	strncpy( nombre, argv[1], 31 );
	i = (int)strtol( argv[2], (char **)NULL, 10);

	if( i<0 || i > 16 ) {
		fprintf( stderr, "El valor de i debe estar en la amplitud [0:15]\n" );
		return 2;
	}

	entropia( nombre, i );

	return 0;
}
